#! /usr/bin/env python

"""
Provides a simple API for streaming data from EDDB in essentially raw format.

Example:
    import misc.eddb
    for sysdata in misc.eddb.SystemsQuery():
        print(sysdata)

Original author: oliver@kfs.org
"""

import json
import sys
import transfers

BASE_URL = "http://eddb.io/archive/v3/"
COMMODITIES_JSON = BASE_URL + "commodities.json"
SYSTEMS_JSON = BASE_URL + "systems.json"
STATIONS_EXT_JSON = BASE_URL + "stations.json"
STATIONS_LITE_JSON = BASE_URL + "stations_lite.json"


class EDDBQuery(object):
    """
    Base class for querying an EDDB data set and converting the
    JSON results into an iterable stream.
    
    Example:
        for entity in EDDBQuery():
            print(entity)
    """
    
    url = None      # Define in derived classes
    
    def __init__(self):
        assert self.url
        self.jsonData = transfers.get_json_data(self.url)
    
    def __iter__(self):
        return iter(self.jsonData)

class CommoditiesQuery(EDDBQuery):
    """
    Streams Commodities data from EDDB.
    
    Example:
        for comm in CommoditiesQuery():
            print(comm['name'])
    """
    url = COMMODITIES_JSON

class SystemsQuery(EDDBQuery):
    """
    Streams System data from EDDB.
    
    Example:
        for system in SystemsQuery():
            print(system['name'])
    """
    url = SYSTEMS_JSON

class StationsQuery(EDDBQuery):
    """
    Streams Station data from EDDB without trade data.
    
    Example:
        for station in StationsQuery():
            print(station['name'])
    """
    url = STATIONS_LITE_JSON

class StationsExtQuery(StationsQuery):
    """
    Streams extended Station data from EDDB with trade data.
    
    Example:
        for station in StationsExtQuery():
            print(station['name'])
    """
    url = STATIONS_EXT_JSON




from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from commands.commandenv import ResultRow
from commands.parsing import *
from formatting import RowFormat, ColumnFormat

######################################################################
# Parser config

help = 'Describe your command briefly here for the top-level --help.'
name = 'TEMPLATE'       # name of your .py file excluding the _cmd
epilog = None           # text to print at the bottom of --help
wantsTradeDB = True     # Should we try to load the cache at startup?
usesTradeData = True    # Will we be needing trading data?
arguments = [
    #ParseArgument('near', help='System to start from', type=str),
]
switches = [
    #ParseArgument('--ly-per',
    #       help='Maximum light years per jump.',
    #       dest='maxLyPer',
    #       metavar='N.NN',
    #       type=float,
    #   ),
]

######################################################################
# Helpers

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    """
    Implement code that validates arguments, collects and prepares
    any data you will need to generate your results for the user.
    
    If your command has finished and has no output to generate,
    return None, otherwise return "results" to be forwarded to
    the 'render' function.
    
    DO NOT print() during 'run', this allows run() functions to
    be re-used between modules and allows them to be used beyond
    the trade.py command line - e.g. someone writing a TD GUI
    will call run() and then render the results themselves.
    """
    
    ### TODO: Implement
    
    return results

######################################################################
# Transform result set into output

def render(results, cmdenv, tdb):
    """
    If run() returns a non-None value, the trade.py code will then
    call the corresponding render() function.
    
    This is where you should generate any output from your command.
    """
    
    ### TODO: Implement


#!/usr/bin/env python3.6

import tradedb
from tradeexcept import TradeException
import json

class UnknownSystemError(TradeException):
    def __str__(self):
        return "Unknown System: " + ' '.join(self.args)

class UnknownStationError(TradeException):
    def __str__(self):
        return "Unknown Station: " + ' '.join(self.args)

def lookup_system(tdb, tdenv, name, x, y, z):
    try:
        system = tdb.systemByName[name.upper()]
    except KeyError:
        system = None
    if not system:
        try:
            system = tdb.lookupSystem(name)
        except LookupError:
            pass
    
    if system:
        if (system.posX != x or system.posY != y or system.posZ != z):
            raise Exception("System {} position mismatch: "
                    "Got {},{},{} expected {},{},{}".format(
                        name,
                        x, y, z,
                        system.posX, system.posY, system.posZ
            ))
        return system
    
    newSystem = "@{} [{}, {}, {}]".format(
            name, x, y, z
    )
    
    candidates = []
    for candidate in tdb.systemByID.values():
        if (candidate.posX == x and
                candidate.posY == y and
                candidate.posZ == z
                ):
            candidates.append(candidate)
    
    if len(candidates) == 1:
        candidate = candidates[0]
        if candidate.casefold() != name.casefold():
            if not tdenv.quiet:
                print("System name mismatch: "
                        "Local: {}, "
                        "Remote: {}, "
                        "Coords: {}, {}, {}".format(
                            name,
                            candidate.dbname,
                            candidate.posX,
                            candidate.posY,
                            candidate.posZ,
                ))
        return candidates[0]
    
    if len(candidates):
        raise Exception("System {} matches co-ordinates for systems: {}" +
                ','.join([system.name for system in candidates])
                )
    
    if tdenv.addUnknown:
        return tdb.addLocalSystem(name, x, y, z)
    
    return None

def lookup_station(
        tdb, tdenv,
        system, name,
        lsFromStar, blackMarket, maxPadSize
        ):
    station = None
    normalizedName = tradedb.TradeDB.normalizedStr(name)
    for stn in system.stations:
        stnNormalizedName = tradedb.TradeDB.normalizedStr(stn.dbname)
        if stnNormalizedName == normalizedName:
            station = stn
            break
    
    if not station:
        if not tdenv.addUnknown:
            return None
        station = tdb.addLocalStation(system, name)
    
    # Now set the parameters
    tdb.updateLocalStation(
            stn, lsFromStar, blackMarket, maxPadSize
    )
    return station

def load_prices_json(
        tdb,
        tdenv,
        jsonText
        ):
    """
    Take data from a prices file and load it into the database.
    """
    
    data = json.loads(jsonText)
    
    sysData = data['sys']
    sysName = sysData['name']
    pos = sysData['pos']
    
    stnData = data['stn']
    stnName = stnData['name']
    lsFromStar = stnData['ls']
    
    try:
        blackMarket = stnData['bm'].upper()
        if not blackMarket in [ 'Y', 'N' ]:
            blackMarket = '?'
    except KeyError:
        blackMarket = '?'
    
    system = lookup_system(
            tdb, tdenv,
            sysName,
            pos[0], pos[1], pos[2],
            )
    if not system:
        if not tdenv.ignoreUnknown:
            raise UnknownSystemError(sysName)
        if not tdenv.quiet:
            print("NOTE: Ignoring unknown system: {} [{},{},{}]".format(
                    sysName,
                    pos[0], pos[1], pos[2],
            ))
        return
    if system.dbname != sysName and tdenv.detail:
        print("NOTE: Treating '{}' as '{}'".format(
                name, system.dbname
        ))
    tdenv.DEBUG1("- System: {}", system.dbname)
    
    station = lookup_station(
            tdb, tdenv,
            system,
            stnName,
            lsFromStar,
            blackMarket,
            )
    if not station:
        if tdenv.ignoreUnknown:
            raise UnknownStationError(stnName)
        if not tdenv.quiet:
            print("NOTE: Ignoring unknown station: {}/{}".format(
                    sysName.upper(), stnName
            ))
        return
    tdenv.DEBUG1("- Station: {}", station.dbname)

def generate_prices_json(
        tdb,
        tdenv,
        station,
        ):
    """
    Generate a JSON dump of the specified station along
    with everything we know about the station and the system
    it is in. 
    
    tdb:
        The TradeDB object to use
    tdenv:
        Settings
    station:
        Station to dump
    """
    
    system = station.system
    
    stationData = {
        'cmdr': tdenv.commander or "unknown",
        'src': 'td/price-json',
        'sys': {
            'name': system.dbname,
            'pos': [ system.posX, system.posY, system.posZ ],
        },
        'stn': {
            'name': station.dbname,
            'ls': station.lsFromStar,
        },
        'items': {}
    }
    
    conn = tdb.getDB()
    cur = conn.cursor()
    cur.execute("""
            SELECT  si.item_id,
                    si.modified,
                    sb.price, sb.units, sb.level,
                    ss.price, ss.units, ss.level
              FROM  StationItem AS si
                    LEFT OUTER JOIN StationBuying AS sb
                        ON (
                            sb.station_id = si.station_id
                            AND sb.item_id = si.item_id
                        )
                    LEFT OUTER JOIN StationSelling AS ss
                        ON (
                            ss.station_id = si.station_id
                            AND ss.item_id = si.item_id
                        )
             WHERE  si.station_id = ?
    """, [station.ID])
    
    items = {}
    
    lastModified = "0"
    for (
            itemID, modified,
            sbPrice, sbUnits, sbLevel,
            ssPrice, ssUnits, ssLevel
    ) in cur:
        lastModified = max(lastModified, modified)
        item = tdb.itemByID[itemID]
        itemData = items[item.dbname] = {
            'm': modified,
        }
        if sbPrice and (sbUnits >= 0 or sbLevel >= 0):
            itemData['b'] = [ sbPrice, sbUnits, sbLevel ]
        elif sbPrice:
            itemData['b'] = sbPrice
        if ssPrice and (ssUnits >= 0 or sbLevel >= 0):
            itemData['s'] = [
                ssPrice, ssUnits, ssLevel
            ]
        elif ssPrice:
            itemData['s'] = ssPrice
    
    # dedupe timestamps.
    for itemData in items.values():
        if itemData['m'] == lastModified:
            del itemData['m']
    
    stationData['m'] = lastModified
    stationData['items'] = items
    
    return json.dumps(stationData, separators=(',',':'))



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

from .exceptions import *
from .parsing import *
from itertools import chain
from pathlib import Path

from .. import cache, plugins, transfers
import math
import re

try:
    import tkinter
    import tkinter.filedialog as tkfd
    hasTkInter = True
except ImportError:
    hasTkInter = False

######################################################################
# Parser config

help=(
    "TD data import system. On its own, this command lets you "
    "merge station prices from a '.prices' file (entries in the "
    "file that are older than your local data are not loaded)."
)
name='import'
epilog=(
    "This sub-command provides a plugin infrastructure, and comes "
    "with a module to import data from Maddavo's Market Share "
    "(http://www.davek.com.au/td/).\n"
    "See \"import -P maddavo -O help\" for more help."
)
wantsTradeDB=False
arguments = [
]
switches = [
    MutuallyExclusiveGroup(
        ParseArgument('filename',
            help='Name of the file to read.',
            type=str,
            default=None,
        ),
        ParseArgument('--plug', '-P',
                help="Use the specified import plugin.",
                type=str,
                default=None,
        ),
    ),
    ParseArgument('--url',
        help='Name of the file to read.',
        type=str,
        default=None,
    ),
    ParseArgument('--download',
        help='Stop after downloading.',
        action='store_true',
        default=False,
    ),
    ParseArgument(
        '--ignore-unknown', '-i',
        default=False, action='store_true',
        dest='ignoreUnknown',
        help=(
            "Data for systems, stations and items that are not "
            "recognized is reported as warning but skipped."
        ),
    ),
    ParseArgument(
        '--option', '-O',
        default=[], action='append',
        dest='pluginOptions',
        help=(
            "Provides a way to pass additional arguments to plugins."
        ),
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--reset-all',
            help='Clear the database before importing.',
            action='store_true',
            default=False,
        ),
        ParseArgument('--merge-import', '-M',
            help=(
                'Merge the import file with the existing local database: '
                'only loads values that have an explicit entry with a '
                'newer timestamp than the existing data. Local values '
                'are only removed if there is an explicit entry with a '
                '0/0 demand/supply price.'
            ),
            action='store_true',
            default=False,
            dest='mergeImport',
        ),
    ),
]

######################################################################
# Helpers

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    # If we're using a plugin, initialize that first.
    if cmdenv.plug:
        if cmdenv.pluginOptions:
            cmdenv.pluginOptions = chain.from_iterable(
                opt.split(',') for opt in cmdenv.pluginOptions
            )
        try:
            pluginClass = plugins.load(cmdenv.plug, "ImportPlugin")
        except plugins.PluginException as e:
            raise CommandLineError("Plugin Error: "+str(e))
        
        # Initialize the plugin
        plugin = pluginClass(tdb, cmdenv)
        
        # Run the plugin. If it returns False, then it did everything
        # that needs doing and we can stop now.
        # If it returns True, it is returning control to the module.
        if not plugin.run():
            return None
    
    tdb.reloadCache()
    tdb.close()
    
    if cmdenv.filename:
        if re.match(r"^https?://", cmdenv.filename, re.IGNORECASE):
            cmdenv.url, cmdenv.filename = cmdenv.filename, None
    
    if cmdenv.url:
        cmdenv.filename = cmdenv.filename or "import.prices"
        transfers.download(cmdenv, cmdenv.url, cmdenv.filename)
        if cmdenv.download:
            return None
    
    # If the filename specified was "-" or None, then go ahead
    # and present the user with an open file dialog.
    if not cmdenv.filename and hasTkInter:
        tk = tkinter.Tk()
        tk.withdraw()
        filetypes = (
                ("TradeDangerous '.prices' Files", "*.prices"),
                ("All Files", "*.*"),
                )
        filename = tkfd.askopenfilename(
                    title="Select the file to import",
                    initialfile="TradeDangerous.prices",
                    filetypes=filetypes,
                    initialdir='.',
                )
        if not filename:
            raise SystemExit("Aborted")
        cmdenv.filename = filename
    
    # check the file exists.
    if cmdenv.filename != "-":
        fh = None
        filePath = Path(cmdenv.filename)
        if not filePath.is_file():
            raise CommandLineError("File not found: {}".format(
                        str(filePath)
                    ))
    else:
        filePath = "stdin"
        fh = sys.stdin
    
    if cmdenv.plug:
        if not plugin.finish():
            cache.regeneratePricesFile()
            return None
    
    cache.importDataFromFile(tdb, cmdenv, filePath, pricesFh=fh, reset=cmdenv.reset)
    
    return None



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .exceptions import PadSizeError, PlanetaryError

######################################################################
# Parsing Helpers

class ParseArgument(object):
    """
        Provides argument forwarding so that 'makeSubParser' can take function-like arguments.
    """
    def __init__(self, *args, **kwargs):
        self.args, self.kwargs = args, kwargs


class MutuallyExclusiveGroup(object):
    def __init__(self, *args):
        self.arguments = list(args)


######################################################################
# Derived Parsers


class CreditParser(int):
    """
    argparse helper for parsing numeric prefixes, i.e.
    'k' for thousand, 'm' for million and 'b' for billion.
    """
    suffixes = {'k': 10**3, 'm': 10**6, 'b': 10**9}
    def __new__(cls, val, **kwargs):
        if isinstance(val, str):
            if val[-1] in CreditParser.suffixes:
                val = int(float(val[:-1]) * CreditParser.suffixes[val[-1]])
        return super().__new__(cls, val, **kwargs)


class PadSizeArgument(int):
    """
    argparse helper for --pad-size
    """
    class PadSizeParser(str):
        def __new__(cls, val, **kwargs):
            if not isinstance(val, str):
                raise PadSizeError(val)
            for v in val:
                if "SML?".find(v.upper()) < 0:
                    raise PadSizeError(val.upper())
            return super().__new__(cls, val, **kwargs)
    
    def __init__(self):
        self.args = ('--pad-size', '-p',)
        self.kwargs = {
            'help': (
                'Limit to stations with one of the specified pad sizes, '
                'e.g. --pad SML? matches any pad, --pad M matches only '
                'medium pads.'
            ),
            'dest': 'padSize',
            'metavar': 'PADSIZES',
            'type': 'padsize',
        }


class AvoidPlacesArgument(ParseArgument):
    def __init__(self):
        self.args = ['--avoid']
        self.kwargs = {
            'action': 'append',
            'help': (
                "Don't list results for the specified systems or stations.\n"
                "Names can be one-per '--avoid' or comma separated, e.g. "
                "'--avoid a,b,c' or '--avoid a,b --avoid c'"
            ),
        }


class SwitchArgument(ParseArgument):
    def __init__(self, help=None):
        if isinstance(self.switches, (tuple, list)):
            self.args = self.switches
        else:
            self.args = (self.switches,)
        help = help or self.help
        self.kwargs = {'action':'store_true', 'dest':self.dest, 'help':help}


class BlackMarketSwitch(SwitchArgument):
    switches = ['--black-market', '--bm']
    dest = 'blackMarket'
    help = 'Require stations known to have a black market.'


class ShipyardSwitch(SwitchArgument):
    switches = ['--shipyard']
    dest = 'shipyard'
    help = 'Require stations known to have a Shipyard.'


class OutfittingSwitch(SwitchArgument):
    switches = ['--outfitting']
    dest = 'outfitting'
    help = 'Require stations known to have Outfitting.'


class RearmSwitch(SwitchArgument):
    switches = ['--rearm']
    dest = 'rearm'
    help = 'Require stations known to sell munitions.'


class RefuelSwitch(SwitchArgument):
    switches = ['--refuel']
    dest = 'refuel'
    help = 'Require stations known to sell fuel.'


class RepairSwitch(SwitchArgument):
    switches = ['--repair']
    dest = 'repair'
    help = 'Require stations known to offer repairs.'


class NoPlanetSwitch(SwitchArgument):
    switches = ['--no-planet']
    dest = 'noPlanet'
    help = 'Require stations to be in space.'


class PlanetaryArgument(int):
    """
    argparse helper for --planetary
    """
    class PlanetaryParser(str):
        def __new__(cls, val, **kwargs):
            if not isinstance(val, str):
                raise PlanetaryError(val)
            for v in val:
                if "YN?".find(v.upper()) < 0:
                    raise PlanetaryError(val.upper())
            return super().__new__(cls, val, **kwargs)
    
    def __init__(self):
        self.args = ['--planetary']
        self.kwargs = {
            'help': (
                'Limit to stations with one of the specified planetary, '
                'e.g. --pla YN? matches any station, --pla Y matches only '
                'planetary stations.'
            ),
            'dest': 'planetary',
            'metavar': 'PLANETARY',
            'type': 'planetary',
        }


__tdParserHelpers = {
    'credits': CreditParser,
    'padsize': PadSizeArgument.PadSizeParser,
    'planetary': PlanetaryArgument.PlanetaryParser,
}

def registerParserHelpers(into):
    for typeName, helper in __tdParserHelpers.items():
        into.register('type', typeName, helper)



from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .commandenv import CommandEnv
from textwrap import TextWrapper

import argparse             # For parsing command line args.
import importlib
import os
import pathlib
import sys

thismodule = sys.modules[__name__]
from . import exceptions
from . import parsing

from . import buildcache_cmd
from . import buy_cmd
from . import export_cmd
from . import import_cmd
from . import local_cmd
from . import market_cmd
from . import nav_cmd
from . import olddata_cmd
from . import rares_cmd
from . import run_cmd
from . import sell_cmd
from . import shipvendor_cmd
from . import station_cmd
from . import trade_cmd
from . import update_cmd

commandIndex = {
    cmd[0:cmd.find('_cmd')]: getattr(thismodule, cmd)
    for cmd in thismodule.__dir__() if cmd.endswith("_cmd")
}

######################################################################
# Helpers

class HelpAction(argparse.Action):
    """
        argparse action helper for printing the argument usage,
        because Python 3.4's argparse is ever-so subtly very broken.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        raise exceptions.UsageError(
                "TradeDangerous help",
                parser.format_help()
        )


def addArguments(group, options, required, topGroup=None):
    """
        Registers a list of options to the specified group. Nodes
        are either an instance of ParseArgument or a list of
        ParseArguments. The list form is considered to be a
        mutually exclusive group of arguments.
    """
    for option in options:
        if isinstance(option, parsing.MutuallyExclusiveGroup):
            exGrp = (topGroup or group).add_mutually_exclusive_group()
            parsing.registerParserHelpers(exGrp)
            addArguments(exGrp, option.arguments, required, topGroup=group)
        else:
            assert not required in option.kwargs
            if option.args[0][0] == '-':
                group.add_argument(*(option.args), required=required, **(option.kwargs))
            else:
                if required:
                    group.add_argument(*(option.args), **(option.kwargs))
                else:
                    group.add_argument(*(option.args), nargs='?', **(option.kwargs))


def _findFromFile(cmd, prefix='.tdrc'):
    if cmd:
        # check the current directory, fall back to home
        filename = '{}_{}'.format(prefix, cmd)
        for dirname in '.', os.path.expanduser('~'):
            cmdPath = pathlib.Path(dirname) / filename
            if cmdPath.exists():
                return cmdPath.resolve()
    return None


class CommandIndex(object):
    def usage(self, argv):
        """
            Generate the outlying usage text for TD.
            This tells the user the list of current
            commands, generated programatically,
            and the outlying command functionality.
        """
        
        text = (
            "Usage: {prog} <command>\n\n"
            "Where <command> is one of:\n\n"
                .format(prog=argv[0])
        )
        
        # Figure out the pre-indentation
        cmdFmt = '  {:<12s}  '
        cmdFmtLen = len(cmdFmt.format(''))
        # Generate a formatter which will produce nicely formatted text
        # that wraps at column 78 but puts continuation text one character
        # indented from where the previous text started, e.g
        #   cmd1    Cmd1 help text stuff
        #            continued cmd1 text
        #   cmd2    Cmd2 help text
        tw = TextWrapper(
                subsequent_indent=' '*(cmdFmtLen + 1),
                width=78,
                drop_whitespace=True,
                expand_tabs=True,
                fix_sentence_endings=True,
                break_long_words=False,
                break_on_hyphens=True,
                )
        
        # List each command with its help text
        lastCmdName = None
        for cmdName, cmd in sorted(commandIndex.items()):
            tw.initial_indent = cmdFmt.format(cmdName)
            text += tw.fill(cmd.help) + "\n"
            lastCmdName = cmdName
        
        # Epilog
        text += (
            "\n"
            "For additional help on a specific command, such as '{cmd}' use\n"
            "  {prog} {cmd} -h"
                .format(prog=argv[0], cmd=lastCmdName)
            )
        return text


    def parse(self, argv, fromfile_prefix='+'):
        if len(argv) <= 1 or argv[1] == '--help' or argv[1] == '-h':
            raise exceptions.UsageError(
                    "TradeDangerous provides a set of trade database "
                    "facilities for Elite:Dangerous.", self.usage(argv))
        
        ### TODO: Break this model up a bit more so that
        ### we just try and import the command you specify,
        ### and only worry about an index when that fails or
        ### the user requests usage.
        cmdName, cmdModule = argv[1].casefold(), None
        try:
            cmdModule = commandIndex[cmdName]
        except KeyError:
            pass
        
        if not cmdModule:
            candidates = []
            for name, module in commandIndex.items():
                if name.startswith(cmdName):
                    candidates.append([name, module])
            if not candidates:
                raise exceptions.CommandLineError(
                        "Unrecognized command, '{}'".format(cmdName),
                        self.usage(argv)
                )
            if len(candidates) > 1:
                raise exceptions.CommandLineError(
                        "Ambiguous command, '{}', "
                        "could match: {}".format(
                            cmdName,
                            ', '.join(c[0] for c in candidates)
                        ),
                        self.usage(argv)
                )
            argv[1] = cmdName = candidates[0][0]
            cmdModule = candidates[0][1]
        
        class ArgParser(argparse.ArgumentParser):
            def error(self, message):
                raise exceptions.CommandLineError(message, self.format_usage())
        
        parser = ArgParser(
                    description="TradeDangerous: "+cmdName,
                    add_help=False,
                    epilog='Use {prog} {cmd} -h for more help'.format(
                            prog=argv[0], cmd=argv[1]
                        ),
                    fromfile_prefix_chars=fromfile_prefix,
                )
        parser.set_defaults(_editing=False)
        parsing.registerParserHelpers(parser)
        
        subParsers = parser.add_subparsers(title='Command Options')
        subParser = subParsers.add_parser(cmdModule.name,
                                    help=cmdModule.help,
                                    add_help=False,
                                    epilog=cmdModule.epilog,
                                    )
        parsing.registerParserHelpers(subParser)
        
        arguments = cmdModule.arguments
        if arguments:
            argParser = subParser.add_argument_group('Required Arguments')
            addArguments(argParser, arguments, True)
        
        switches = cmdModule.switches
        if switches:
            switchParser = subParser.add_argument_group('Optional Switches')
            addArguments(switchParser, switches, False)
        
        # Arguments common to all subparsers.
        stdArgs = subParser.add_argument_group('Common Switches')
        stdArgs.add_argument('-h', '--help',
                    help='Show this help message and exit.',
                    action=HelpAction, nargs=0,
                )
        stdArgs.add_argument('--debug', '-w',
                    help='Enable/raise level of diagnostic output.',
                    default=0, required=False, action='count',
                )
        stdArgs.add_argument('--detail', '-v',
                    help='Increase level  of detail in output.',
                    default=0,required=False, action='count',
                )
        stdArgs.add_argument('--quiet', '-q',
                    help='Reduce level of detail in output.',
                    default=0, required=False, action='count',
                )
        stdArgs.add_argument('--db',
                    help='Specify location of the SQLite database.',
                    default=None, dest='dbFilename', type=str,
                )
        stdArgs.add_argument('--cwd', '-C',
                    help='Change the working directory file accesses are made from.',
                    type=str, required=False,
                )
        stdArgs.add_argument('--link-ly', '-L',
                    help='Maximum lightyears between systems to be considered linked.',
                    type=float,
                    default=None, dest='maxSystemLinkLy',
                )
        
        fromfilePath = _findFromFile(cmdModule.name)
        if fromfilePath:
            argv.insert(2, '{}{}'.format(fromfile_prefix, fromfilePath))
        properties = parser.parse_args(argv[1:])
        
        parsed = CommandEnv(properties, argv, cmdModule)
        parsed.DEBUG0("Command line was: {}", argv)
        
        return parsed



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Jonathan 'eyeonus' Jones 2018
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Modules :: Cache loader
#
#  TD works primarily from an SQLite3 database, but the data in that
#  is sourced from text files.
#   data/TradeDangerous.sql contains the less volatile data - systems,
#   ships, etc
#   data/TradeDangerous.prices contains a description of the price
#   database that is intended to be easily editable and commitable to
#   a source repository.
#
#  TODO: Split prices into per-system or per-station files so that
#  we can tell how old data for a specific system is.

from collections import namedtuple
from pathlib import Path
from .tradeexcept import TradeException

from . import corrections, utils
import csv
import math
import os
from . import prices
import re
import sqlite3
import sys

######################################################################
# Regular expression patterns. Here be draegons.
# If you add new patterns:
# - use fragments and re.VERBOSE (see itemPriceRe)
# - use named captures (?P<name> ...)
# - include comments

# # Match the '@ SYSTEM/Station' line
systemStationRe = re.compile(r'^\@\s*(.*)/(.*)')

# # Price Line matching

# first part of any prices line is the item name and paying/asking price
itemPriceFrag = r"""
    # match item name, allowing spaces in the name
    (?P<item> .*?)
\s+
    # price station is buying the item for
    (?P<sell> \d+)
\s+
    # price station is selling item for
    (?P<buy> \d+)
"""

# time formats per https://www.sqlite.org/lang_datefunc.html
# YYYY-MM-DD HH:MM:SS
# YYYY-MM-DDTHH:MM:SS
# HH:MM:SS
# 'now'
timeFrag = r'(?P<time>(\d{4}-\d{2}-\d{2}[T ])?\d{2}:\d{2}:\d{2}|now)'

# <name> <sell> <buy> [ <demand> <supply> [ <time> | now ] ]
qtyLevelFrag = r"""
    unk             # You can just write 'unknown'
|   \?              # alias for unknown
|   n/a             # alias for 0L0
|   -               # alias for 0L0
|   \d+[\?LMH]      # Or <number><level> where level is L(ow), M(ed) or H(igh)
|   0               # alias for n/a
|   bug
"""
newItemPriceRe = re.compile(r"""
^
    {base_f}
    (
    \s+
        # demand units and level
        (?P<demand> {qtylvl_f})
    \s+
        # supply units and level
        (?P<supply> {qtylvl_f})
        # time is optional
        (?:
        \s+
            {time_f}
        )?
    )?
\s*
$
""".format(base_f = itemPriceFrag, qtylvl_f = qtyLevelFrag, time_f = timeFrag),
            re.IGNORECASE + re.VERBOSE)

######################################################################
# Exception classes


class BuildCacheBaseException(TradeException):
    """
    Baseclass for BuildCache exceptions
    Attributes:
        fileName    Name of file being processedStations
        lineNo      Line the error occurred on
        error       Description of the error
    """

    def __init__(self, fromFile, lineNo, error = None):
        self.fileName = fromFile.name
        self.lineNo = lineNo
        self.category = "ERROR"
        self.error = error or "UNKNOWN ERROR"
    
    def __str__(self):
        return "{}:{} {} {}".format(
                self.fileName, self.lineNo,
                self.category,
                self.error,
        )


class UnknownSystemError(BuildCacheBaseException):
    """
    Raised when the file contains an unknown star name.
    """

    def __init__(self, fromFile, lineNo, key):
        error = 'Unrecognized SYSTEM: "{}"'.format(key)
        super().__init__(fromFile, lineNo, error)


class UnknownStationError(BuildCacheBaseException):
    """
    Raised when the file contains an unknown star/station name.
    """

    def __init__(self, fromFile, lineNo, key):
        error = 'Unrecognized STAR/Station: "{}"'.format(key)
        super().__init__(fromFile, lineNo, error)


class UnknownItemError(BuildCacheBaseException):
    """
    Raised in the case of an item name that we don't know.
    Attributes:
        itemName   Key we tried to look up.
    """

    def __init__(self, fromFile, lineNo, itemName):
        error = 'Unrecognized item name: "{}"'.format(itemName)
        super().__init__(fromFile, lineNo, error)


class DuplicateKeyError(BuildCacheBaseException):
    """
        Raised when an item is being redefined.
    """

    def __init__(self, fromFile, lineNo, keyType, keyValue, prevLineNo):
        super().__init__(fromFile, lineNo,
                "Second occurrance of {keytype} \"{keyval}\", "
                "previous entry at line {prev}.".format(
                    keytype = keyType,
                    keyval = keyValue,
                    prev = prevLineNo
                ))


class DeletedKeyError(BuildCacheBaseException):
    """
    Raised when a key value in a .csv file is marked as DELETED in the
    corrections file.
    """

    def __init__(self, fromFile, lineNo, keyType, keyValue):
        super().__init__(fromFile, lineNo,
                "{} '{}' is marked as DELETED and should not be used.".format(
                    keyType, keyValue
        ))


class DeprecatedKeyError(BuildCacheBaseException):
    """
    Raised when a key value in a .csv file has a correction; the old
    name should not appear in the .csv file.
    """

    def __init__(self, fromFile, lineNo, keyType, keyValue, newValue):
        super().__init__(fromFile, lineNo,
                "{} '{}' is deprecated "
                "and should be replaced with '{}'.".format(
                    keyType, keyValue, newValue
        ))


class MultipleStationEntriesError(DuplicateKeyError):
    """ Raised when a station appears multiple times in the same file. """

    def __init__(self, fromFile, lineNo, facility, prevLineNo):
        super().__init__(fromFile, lineNo, 'station', facility, prevLineNo)


class MultipleItemEntriesError(DuplicateKeyError):
    """ Raised when one item appears multiple times in the same station. """

    def __init__(self, fromFile, lineNo, item, prevLineNo):
        super().__init__(fromFile, lineNo, 'item', item, prevLineNo)


class SyntaxError(BuildCacheBaseException):
    """
    Raised when an invalid line is read.
    Attributes:
        problem     The problem that occurred
        text        Offending text
    """

    def __init__(self, fromFile, lineNo, problem, text):
        error = "{},\ngot: '{}'.".format(problem, text.strip())
        super().__init__(fromFile, lineNo, error)


class SupplyError(BuildCacheBaseException):
    """
    Raised when a supply field is incorrectly formatted.
    """

    def __init__(self, fromFile, lineNo, category, problem, value):
        error = "Invalid {} supply value: {}. Got: {}". \
                    format(category, problem, value)
        super().__init__(fromFile, lineNo, error)

######################################################################
# Helpers


def parseSupply(pricesFile, lineNo, category, reading):
    units, level = reading[0:-1], reading[-1]
    levelNo = "??LMH".find(level.upper()) - 1
    if levelNo < -1:
        raise SupplyError(
            pricesFile, lineNo, category, reading,
            'Unrecognized level suffix: "{}": '
            "expected one of 'L', 'M', 'H' or '?'".format(
                level
            )
        )
    try:
        unitsNo = int(units)
        if unitsNo < 0:
            raise ValueError("unsigned unit count")
        if unitsNo == 0:
            return 0, 0
        return unitsNo, levelNo
    except ValueError:
        pass
    
    raise SupplyError(
        pricesFile, lineNo, category, reading,
        'Unrecognized units/level value: "{}": '
        "expected '-', '?', or a number followed "
        "by a level (L, M, H or ?).".format(
            level
        )
    )

######################################################################
# Code
######################################################################


def getSystemByNameIndex(cur):
    """ Build station index in STAR/Station notation """
    cur.execute("""
            SELECT system_id, UPPER(system.name)
              FROM System
        """)
    return { name: ID for (ID, name) in cur }


def getStationByNameIndex(cur):
    """ Build station index in STAR/Station notation """
    cur.execute("""
            SELECT station_id,
                    UPPER(system.name) || '/' || UPPER(station.name)
              FROM System
                   INNER JOIN Station
                      USING (system_id)
        """)
    return { name: ID for (ID, name) in cur }


def getItemByNameIndex(cur):
    """
        Generate item name index.
    """
    cur.execute("SELECT item_id, UPPER(name) FROM item")
    return { name: itemID for (itemID, name) in cur }


def processPrices(tdenv, priceFile, db, defaultZero):
    """
        Yields SQL for populating the database with prices
        by reading the file handle for price lines.
    """
    
    DEBUG0, DEBUG1 = tdenv.DEBUG0, tdenv.DEBUG1
    DEBUG0("Processing prices file: {}", priceFile)
    
    cur = db.cursor()
    ignoreUnknown = tdenv.ignoreUnknown
    quiet = tdenv.quiet
    merging = tdenv.mergeImport
    
    systemByName = getSystemByNameIndex(cur)
    stationByName = getStationByNameIndex(cur)
    stationByName.update(
        (sys, ID)
        for sys, ID in corrections.stations.items()
        if isinstance(ID, int)
    )
    sysCorrections = corrections.systems
    stnCorrections = {
        stn: alt
        for stn, alt in corrections.stations.items()
        if isinstance(alt, str)
    }
    
    itemByName = getItemByNameIndex(cur)
    
    defaultUnits = -1 if not defaultZero else 0
    defaultLevel = -1 if not defaultZero else 0
    
    stationID = None
    facility = None
    processedStations = {}
    processedSystems = set()
    processedItems = {}
    stationItemDates = {}
    itemPrefix = ""
    DELETED = corrections.DELETED
    items, zeros, buys, sells = [], [], [], []
    
    lineNo, localAdd = 0, 0
    if not ignoreUnknown:

        def ignoreOrWarn(error):
            raise error

    elif not quiet:
        ignoreOrWarn = tdenv.WARN
    
    def changeStation(matches):
        nonlocal facility, stationID
        nonlocal processedStations, processedItems, localAdd
        nonlocal stationItemDates
        
        # ## Change current station
        stationItemDates = {}
        systemNameIn, stationNameIn = matches.group(1, 2)
        systemName, stationName = systemNameIn.upper(), stationNameIn.upper()
        corrected = False
        facility = "/".join((systemName, stationName))
        
        # Make sure it's valid.
        stationID = DELETED
        newID = stationByName.get(facility, -1)
        DEBUG0("Selected station: {}, ID={}", facility, newID)
        if newID is DELETED:
            DEBUG1("DELETED Station: {}", facility)
            return
        if newID < 0:
            if utils.checkForOcrDerp(tdenv, systemName, stationName):
                return
            corrected = True
            altName = sysCorrections.get(systemName, None)
            if altName is DELETED:
                DEBUG1("DELETED System: {}", facility)
                return
            if altName:
                DEBUG1("SYSTEM '{}' renamed '{}'", systemName, altName)
                systemName, facility = altName, "/".join((altName, stationName))
            
            systemID = systemByName.get(systemName, -1)
            if systemID < 0:
                ignoreOrWarn(
                    UnknownSystemError(priceFile, lineNo, facility)
                )
                return
            
            altStation = stnCorrections.get(facility, None)
            if altStation is DELETED:
                DEBUG1("DELETED Station: {}", facility)
                return
            if altStation:
                DEBUG1("Station '{}' renamed '{}'", facility, altStation)
                stationName = altStation.upper()
                facility = "/".join((systemName, stationName))
            
            newID = stationByName.get(facility, -1)
            if newID is DELETED:
                DEBUG1("Renamed station DELETED: {}", facility)
                return
        
        if newID < 0:
            if not ignoreUnknown:
                ignoreOrWarn(
                        UnknownStationError(priceFile, lineNo, facility)
                )
                return
            name = utils.titleFixup(stationName)
            inscur = db.cursor()
            inscur.execute("""
                INSERT INTO Station (
                    system_id, name,
                    ls_from_star,
                    blackmarket,
                    max_pad_size,
                    market,
                    shipyard,
                    modified
                ) VALUES (
                    ?, ?, 0, '?', '?', '?', '?',
                    DATETIME('now')
                )
            """, [systemID, name])
            newID = inscur.lastrowid
            stationByName[facility] = newID
            tdenv.NOTE("Added local station placeholder for {} (#{})",
                    facility, newID
            )
            localAdd += 1
        elif newID in processedStations:
            # Check for duplicates
            if not corrected:
                raise MultipleStationEntriesError(
                    priceFile, lineNo, facility,
                    processedStations[newID]
                )
        
        stationID = newID
        processedSystems.add(systemName)
        processedStations[stationID] = lineNo
        processedItems = {}
        
        cur = db.execute("""
            SELECT item_id, modified
              FROM StationItem
             WHERE station_id = ?
        """, [stationID])
        stationItemDates = {ID: modified for ID, modified in cur}
    
    addItem, addZero = items.append, zeros.append
    getItemID = itemByName.get
    newItems, updtItems, ignItems = 0, 0, 0
    
    def processItemLine(matches):
        nonlocal newItems, updtItems, ignItems
        itemName, modified = matches.group('item', 'time')
        itemName = itemName.upper()
        
        # Look up the item ID.
        itemID = getItemID(itemName, -1)
        if itemID < 0:
            oldName = itemName
            itemName = corrections.correctItem(itemName)
            if itemName == DELETED:
                DEBUG1("DELETED {}", oldName)
                return
            itemName = itemName.upper()
            itemID = getItemID(itemName, -1)
            if itemID < 0:
                ignoreOrWarn(
                    UnknownItemError(priceFile, lineNo, itemName)
                )
                return
            DEBUG1("Renamed {} -> {}", oldName, itemName)
        
        lastModified = stationItemDates.get(itemID, None)
        if lastModified and merging:
            if modified and modified != 'now' and modified <= lastModified:
                DEBUG1("Ignoring {} @ {}: {} <= {}".format(
                    itemName, facility,
                    modified, lastModified,
                ))
                if modified < lastModified:
                    ignItems += 1
                return
        
        # Check for duplicate items within the station.
        if itemID in processedItems:
            ignoreOrWarn(
                MultipleItemEntriesError(
                    priceFile, lineNo,
                    "{}".format(itemName),
                    processedItems[itemID]
                )
            )
            return
        
        demandCr, supplyCr = matches.group('sell', 'buy')
        demandCr, supplyCr = int(demandCr), int(supplyCr)
        demandString, supplyString = matches.group('demand', 'supply')
        
        if demandCr == 0 and supplyCr == 0:
            if lastModified:
                addZero((stationID, itemID))
        else:
            if lastModified:
                updtItems += 1
            else:
                newItems += 1
            if demandString:
                if demandString == "?":
                    demandUnits, demandLevel = -1, -1
                elif demandString == "-":
                    demandUnits, demandLevel = 0, 0
                else:
                    demandUnits, demandLevel = parseSupply(
                        priceFile, lineNo, 'demand', demandString
                    )
            else:
                demandUnits, demandLevel = defaultUnits, defaultLevel
            
            if demandString and supplyString:
                if supplyString == "?":
                    supplyUnits, supplyLevel = -1, -1
                elif supplyString == "-":
                    supplyUnits, supplyLevel = 0, 0
                else:
                    supplyUnits, supplyLevel = parseSupply(
                        priceFile, lineNo, 'supply', supplyString
                    )
            else:
                supplyUnits, supplyLevel = defaultUnits, defaultLevel
            
            if modified == 'now':
                modified = None  # Use CURRENT_FILESTAMP
            
            addItem((
                stationID, itemID, modified,
                demandCr, demandUnits, demandLevel,
                supplyCr, supplyUnits, supplyLevel,
            ))
        
        processedItems[itemID] = lineNo
    
    for line in priceFile:
        lineNo += 1
        text, _, comment = line.partition('#')
        text = text.strip()
        if not text:
            continue
        
        # replace whitespace with single spaces
        if text.find("  "):
            # http://stackoverflow.com/questions/2077897
            text = ' '.join(text.split())
        
        ########################################
        # ## "@ STAR/Station" lines.
        if text.startswith('@'):
            matches = systemStationRe.match(text)
            if not matches:
                raise SyntaxError("Unrecognized '@' line: {}".format(# pylint: disable=no-value-for-parameter
                            text
                        ))
            changeStation(matches)
            continue
        
        if not stationID:
            # Need a station to process any other type of line.
            raise SyntaxError(priceFile, lineNo,
                                "Expecting '@ SYSTEM / Station' line", text)
        if stationID == DELETED:
            # Ignore all values from a deleted station/system.
            continue
        
        ########################################
        # ## "+ Category" lines
        if text.startswith('+'):
            # we now ignore these.
            continue
        
        ########################################
        # ## "Item sell buy ..." lines.
        matches = newItemPriceRe.match(text)
        if not matches:
            raise SyntaxError(priceFile, lineNo,
                        "Unrecognized line/syntax", text)
        
        processItemLine(matches)
    
    numSys = len(processedSystems)
    
    if localAdd > 0:
        tdenv.NOTE(
            "Placeholder stations are added to the local DB only "
            "(not the .CSV).\n"
            "Use 'trade.py export --table Station' "
            "if you /need/ to persist them."
        )
    
    stations = tuple((ID,) for ID in processedStations.keys())
    return stations, items, zeros, newItems, updtItems, ignItems, numSys

######################################################################


def processPricesFile(tdenv, db, pricesPath, pricesFh = None, defaultZero = False):
    tdenv.DEBUG0("Processing Prices file '{}'", pricesPath)
    
    with pricesFh or pricesPath.open('rU', encoding = 'utf-8') as pricesFh:
        stations, items, zeros, newItems, updtItems, ignItems, numSys = processPrices(
            tdenv, pricesFh, db, defaultZero
        )
    
    if not tdenv.mergeImport:
        db.executemany("""
            DELETE FROM StationItem
             WHERE station_id = ?
        """, stations)
    if zeros:
        db.executemany("""
            DELETE FROM StationItem
             WHERE station_id = ?
               AND item_id = ?
        """, zeros)
    removedItems = len(zeros)
    
    if items:
        db.executemany("""
            INSERT OR REPLACE INTO StationItem (
                station_id, item_id, modified,
                demand_price, demand_units, demand_level,
                supply_price, supply_units, supply_level
            ) VALUES (
                ?, ?, IFNULL(?, CURRENT_TIMESTAMP),
                ?, ?, ?,
                ?, ?, ?
            )
        """, items)
    updatedItems = len(items)
    
    tdenv.DEBUG0("Marking populated stations as having a market")
    db.execute(
        "UPDATE Station SET market = 'Y'"
        " WHERE EXISTS"
            " (SELECT station_id FROM StationItem"
              " WHERE StationItem.station_id = Station.station_id"
             ")"
    )
    
    db.commit()
    
    changes = " and ".join("{} {}".format(v, k) for k, v in {
        "new": newItems,
        "updated": updtItems,
        "removed": removedItems,
    }.items() if v) or "0"
    
    tdenv.NOTE(
        "Import complete: "
            "{:s} items "
            "over {:n} stations "
            "in {:n} systems",
                changes,
                len(stations),
                numSys,
    )
    
    if ignItems:
        tdenv.NOTE("Ignored {} items with old data", ignItems)

######################################################################


def depCheck(importPath, lineNo, depType, key, correctKey):
    if correctKey == key:
        return
    if correctKey == corrections.DELETED:
        raise DeletedKeyError(importPath, lineNo, depType, key)
    raise DeprecatedKeyError(importPath, lineNo, depType, key, correctKey)


def deprecationCheckSystem(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'System',
        line[0], corrections.correctSystem(line[0]),
    )


def deprecationCheckStation(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'System',
        line[0], corrections.correctSystem(line[0]),
    )
    depCheck(
        importPath, lineNo, 'Station',
        line[1], corrections.correctStation(line[0], line[1]),
    )


def deprecationCheckCategory(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'Category',
        line[0], corrections.correctCategory(line[0]),
    )


def deprecationCheckItem(importPath, lineNo, line):
    depCheck(
        importPath, lineNo, 'Category',
        line[0], corrections.correctCategory(line[0]),
    )
    depCheck(
        importPath, lineNo, 'Item',
        line[1], corrections.correctItem(line[1]),
    )


def processImportFile(tdenv, db, importPath, tableName):
    tdenv.DEBUG0(
        "Processing import file '{}' for table '{}'",
        str(importPath), tableName
    )
    
    fkeySelectStr = ("("
            "SELECT {newValue}"
            " FROM {table}"
            " WHERE {stmt}"
            ")"
    )
    uniquePfx = "unq:"
    uniqueLen = len(uniquePfx)
    ignorePfx = "!"
    
    with importPath.open('rU', encoding = 'utf-8') as importFile:
        csvin = csv.reader(
            importFile, delimiter = ',', quotechar = "'", doublequote = True
        )
        # first line must be the column names
        columnDefs = next(csvin)
        columnCount = len(columnDefs)
        
        # split up columns and values
        # this is necessqary because the insert might use a foreign key
        bindColumns = []
        bindValues = []
        joinHelper = []
        uniqueIndexes = []
        for (cIndex, cName) in enumerate(columnDefs):
            colName, _, srcKey = cName.partition('@')
            # is this a unique index?
            if colName.startswith(uniquePfx):
                uniqueIndexes.append(cIndex)
                colName = colName[uniqueLen:]
            if not srcKey:
                # no foreign key, straight insert
                bindColumns.append(colName)
                bindValues.append('?')
                continue
            
            queryTab, _, queryCol = srcKey.partition('.')
            if colName.startswith(ignorePfx):
                # this column is only used to resolve an FK
                assert srcKey
                colName = colName[len(ignorePfx):]
                joinHelper.append((colName, queryTab, queryCol))
                continue
            
            # foreign key, we need to make a select
            joinTable = [ queryTab ]
            joinStmt = []
            for nextCol, nextTab, nextJoin in joinHelper:
                joinTable.append(
                    "INNER JOIN {} USING({})".format(nextTab, nextJoin)
                )
                joinStmt.append(
                    "{}.{} = ?".format(nextTab, nextCol)
                )
            joinHelper = []
            joinStmt.append("{}.{} = ?".format(queryTab, colName))
            bindColumns.append(queryCol)
            bindValues.append(
                fkeySelectStr.format(
                    newValue = srcKey,
                    table = " ".join(joinTable),
                    stmt = " AND ".join(joinStmt),
                )
            )
        # now we can make the sql statement
        sql_stmt = """
            INSERT OR REPLACE INTO {table} ({columns}) VALUES({values})
        """.format(
                table = tableName,
                columns = ','.join(bindColumns),
                values = ','.join(bindValues)
            )
        tdenv.DEBUG0("SQL-Statement: {}", sql_stmt)
        
        # Check if there is a deprecation check for this table.
        deprecationFn = getattr(
            sys.modules[__name__],
            "deprecationCheck" + tableName,
            None
        )
        
        # import the data
        importCount = 0
        uniqueIndex = dict()
        
        for linein in csvin:
            if not linein:
                continue
            lineNo = csvin.line_num
            if len(linein) == columnCount:
                tdenv.DEBUG1("       Values: {}", ', '.join(linein))
                if deprecationFn:
                    try:
                        deprecationFn(importPath, lineNo, linein)
                    except (DeprecatedKeyError, DeletedKeyError) as e:
                        if not tdenv.ignoreUnknown:
                            raise e
                        e.category = "WARNING"
                        tdenv.NOTE("{}", e)
                        continue
                if uniqueIndexes:
                    # Need to construct the actual unique index key as
                    # something less likely to collide with manmade
                    # values when it's a compound.
                    keyValues = [
                        str(linein[col]).upper()
                        for col in uniqueIndexes
                    ]
                    key = ":!:".join(keyValues)
                    prevLineNo = uniqueIndex.get(key, 0)
                    if prevLineNo:
                        # Make a human-readable key
                        key = "/".join(keyValues)
                        raise DuplicateKeyError(
                            importPath, lineNo,
                            "entry", key,
                            prevLineNo
                        )
                    uniqueIndex[key] = lineNo
                
                try:
                    db.execute(sql_stmt, linein)
                except Exception as e:
                    raise SystemExit(
                        "*** INTERNAL ERROR: {err}\n"
                        "CSV File: {file}:{line}\n"
                        "SQL Query: {query}\n"
                        "Params: {params}\n"
                        .format(
                            err = str(e),
                            file = str(importPath),
                            line = lineNo,
                            query = sql_stmt.strip(),
                            params = linein
                        )
                    ) from None
                importCount += 1
            else:
                tdenv.NOTE(
                        "Wrong number of columns ({}:{}): {}",
                            importPath,
                            lineNo,
                            ', '.join(linein)
                )
        db.commit()
        tdenv.DEBUG0("{count} {table}s imported",
                            count = importCount,
                            table = tableName)

######################################################################


def buildCache(tdb, tdenv):
    """
    Rebuilds the SQlite database from source files.
    
    TD's data is either "stable" - information that rarely changes like Ship
    details, star systems etc - and "volatile" - pricing information, etc.
    
    The stable data starts out in data/TradeDangerous.sql while other data
    is stored in custom-formatted text files, e.g. ./TradeDangerous.prices.
    
    We load both sets of data into an SQLite database, after which we can
    avoid the text-processing overhead by simply checking if the text files
    are newer than the database.
    """
    
    tdenv.NOTE(
        "Rebuilding cache file: this may take a few moments.",
        file = sys.stderr
    )
    
    dbPath = tdb.dbPath
    sqlPath = tdb.sqlPath
    pricesPath = tdb.pricesPath
    
    # Create an in-memory database to populate with our data.
    tempPath = dbPath.with_suffix(".new")
    backupPath = dbPath.with_suffix(".old")
    
    if tempPath.exists():
        tempPath.unlink()
    
    tempDB = sqlite3.connect(str(tempPath))
    tempDB.execute("PRAGMA foreign_keys=ON")
    # Read the SQL script so we are ready to populate structure, etc.
    tdenv.DEBUG0("Executing SQL Script '{}' from '{}'", sqlPath, os.getcwd())
    with sqlPath.open('rU', encoding = 'utf-8') as sqlFile:
        sqlScript = sqlFile.read()
        tempDB.executescript(sqlScript)
    
    # import standard tables
    for (importName, importTable) in tdb.importTables:
        try:
            processImportFile(tdenv, tempDB, Path(importName), importTable)
        except FileNotFoundError:
            tdenv.DEBUG0(
                "WARNING: processImportFile found no {} file", importName
            )
        except StopIteration:
            tdenv.NOTE(
                "{} exists but is empty. "
                "Remove it or add the column definition line.",
                importName
            )
    
    # Parse the prices file
    if pricesPath.exists():
        processPricesFile(tdenv, tempDB, pricesPath)
    else:
        tdenv.NOTE(
                "Missing \"{}\" file - no price data.",
                    pricesPath,
                    file = sys.stderr,
        )
    
    tempDB.commit()
    tempDB.close()
    
    tdenv.DEBUG0("Swapping out db files")
    
    if dbPath.exists():
        if backupPath.exists():
            backupPath.unlink()
        dbPath.rename(backupPath)
    tempPath.rename(dbPath)
    
    tdenv.DEBUG0("Finished")

######################################################################


def regeneratePricesFile(tdb, tdenv):
    tdenv.DEBUG0("Regenerating .prices file")
    
    with tdb.pricesPath.open("w", encoding = 'utf-8') as pricesFile:
        prices.dumpPrices(
                tdb.dbFilename,
                prices.Element.full,
                file = pricesFile,
                debug = tdenv.debug)
    
    # Update the DB file so we don't regenerate it.
    os.utime(tdb.dbFilename)

######################################################################


def importDataFromFile(tdb, tdenv, path, pricesFh = None, reset = False):
    """
        Import price data from a file on a per-station basis,
        that is when a new station is encountered, delete any
        existing records for that station in the database.
    """
    
    if not pricesFh and not path.exists():
        raise TradeException("No such file: {}".format(
                    str(path)
                ))
    
    if reset:
        tdenv.DEBUG0("Resetting price data")
        with tdb.getDB() as db:
            db.execute("DELETE FROM StationItem")
            db.commit()
    
    tdenv.DEBUG0("Importing data from {}".format(str(path)))
    processPricesFile(tdenv,
            db = tdb.getDB(),
            pricesPath = path,
            pricesFh = pricesFh,
            )
    
    # If everything worked, we may need to re-build the prices file.
    if path != tdb.pricesPath:
        regeneratePricesFile(tdb, tdenv)



#!/usr/bin/env python

import tradedb
tdb = tradedb.TradeDB()

names = set()

for sys in tdb.systemByID.values():
    for stn in sys.stations:
        names.add(stn.name().upper())

mutators = {
        'D': [ 'O', '0', ],
        'W': [ 'VV', ],
        'R': [ 'IT' ],
        'L': [ 'II' ],
        '&': [ '6' ],
}

def mutate(text, pos):
    for i in range(pos, len(text)):
        char = text[i]
        if char not in mutators:
            continue
        bef, aft = text[:i], text[i+1:]
        for mutant in mutators[char]:
            t2 = bef + mutant + aft
            yield t2
            yield from mutate(str(t2), i+len(mutant))

for name in names:
    for mutant in mutate(name, 0):
        if mutant in names:
            print("{} <-> {}".format(name, mutant))




from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .commandenv import ResultRow
from .exceptions import CommandLineError
from .parsing import *
from ..tradedb import AmbiguityError
from ..tradedb import System, Station
from ..tradedb import TradeDB
from .. import utils
from ..formatting import max_len

from .. import cache
from .. import csvexport
import difflib
import re
import sys

######################################################################
# Parser config

help='Add (or update) a station entry'
name='station'
epilog=None
arguments = [
    ParseArgument(
        'station',
        help='Specify the full name of the station (SYS NAME/STN NAME is also supported).',
        type=str,
        nargs='+',
    ),
]
switches = [
    MutuallyExclusiveGroup(
        ParseArgument(
            '--update', '-u',
            help='Indicates you expect the entry to already exist.',
            action='store_true',
        ),
        ParseArgument(
            '--remove', '-rm',
            help='Indicates you want to remove an entry.',
            action='store_true',
        ),
        ParseArgument(
            '--add', '-a',
            help='Indicates you want to add a new station.',
            action='store_true',
        ),
    ),
    ParseArgument(
        '--ls-from-star',
        help='Number of light seconds between station and star.',
        type=int,
        dest='lsFromStar',
    ),
    # Note: these are not the usual arguments, they're asking the
    # user to assign rather than select values.
    ParseArgument(
        '--black-market', '--bm',
        help='Does the station have a black market (Y or N) or ? if unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
        dest='blackMarket',
    ),
    ParseArgument(
        '--market',
        help='Does the station have a commodities market (Y or N), ? for unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
    ),
    ParseArgument(
        '--shipyard',
        help='Does the station have a shipyard (Y or N) or ? if unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
    ),
    ParseArgument(
        '--pad-size',
        help='Maximum supported pad size (S, M, or L) or ? if unknown.',
        choices=['S', 's', 'M', 'm', 'L', 'l', '?'],
        dest='padSize',
    ),
    ParseArgument(
        '--outfitting',
        help='Does the station provide outfitting (Y or N) or ? if unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
    ),
    ParseArgument(
        '--rearm', '--arm',
        help='Does the station provide rearming (Y or N) or ? if unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
    ),
    ParseArgument(
        '--refuel',
        help='Does the station provide refueling (Y or N) or ? if unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
    ),
    ParseArgument(
        '--repair',
        help='Does the station provide repairs (Y or N) or ? if unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
    ),
    ParseArgument(
        '--planetary',
        help='Is the station on a planet (Y or N) or ? if unknown.',
        choices=['Y', 'y', 'N', 'n', '?'],
    ),
    ParseArgument(
        '--confirm',
        help='For confirmation suspicious looking station names.',
        metavar='CONFIRMATION CODE',
        type=str,
    ),
    ParseArgument(
        '--no-export',
        help='Do not update the .csv files.',
        action='store_true',
        dest='noExport',
    )
]

######################################################################
# Helpers

def makeConfirmationCode(base, candidates):
    """
    Makes a four-digit hex checksum of a station list.
    """
    checksum = base
    for cand in candidates:
        for character in cand:
            checksum <<= 4
            checksum += ord(character)
    
    # python integers don't overflow, so we only need
    # to modulo at the end of the checksum.
    checksum %= 65521       # arbitrary prime < 2^32
    
    return hex(checksum).upper()[2:]


def checkStationDoesNotExist(tdb, cmdenv, system, stationName):
    if not system.stations:
        return
    
    upperName = stationName.upper()
    similarities = set()
    try:
        station = tdb.lookupStation(stationName, system)
        if station.dbname.upper() == upperName:
            raise CommandLineError(
                    "Station \"{}\" "
                    "in system \"{}\" "
                    "already exists.".format(
                        stationName, system.name(),
            ))
        similarities.add(station.dbname.upper())
    except LookupError:
        pass
    except AmbiguityError as e:
        for cand in e.anyMatch:
            similarities.add(e.key(cand).upper())
    
    # Check to see if there are stations with somewhat
    # similar names, but allow the user to get around
    # cases where difflib matches 'X Port' to 'Y Port'.
    stationNames = [
        stn.dbname.upper()
        for stn in system.stations
    ]
    cmdenv.DEBUG0("Comparing {} to {}".format(
            upperName, list(stationNames),
    ))
    candidates = difflib.get_close_matches(
            upperName, stationNames, cutoff=0.6,
    )
    for cand in candidates:
        similarities.add(cand)
    
    if not similarities:
        return
    
    confCode = makeConfirmationCode(system.ID, similarities)
    
    if not cmdenv.confirm:
        raise CommandLineError(
                "\"{}\" contains similar station names:\n"
                "  {}\n"
                "\n"
                "If you want to add this station anyway, re-run the "
                "command and add:\n"
                "  --conf {}".format(
                    system.name(),
                    ', '.join(candidates),
                    confCode
        ))
    
    if cmdenv.confirm.upper() != confCode:
        raise CommandLineError(
            "Wrong confirmation code."
        )
    
    cmdenv.NOTE("Confirmation code accepted.")


def checkSystemAndStation(tdb, cmdenv):
    # In add mode, the user has to be more specific.
    stnName = ' '.join(cmdenv.station).strip()
    
    if not cmdenv.add:
        try:
            station = tdb.lookupPlace(stnName)
        except LookupError:
            raise CommandLineError("Unrecognized Station: {}".format(
                cmdenv.station
            ))
        if not isinstance(station, Station):
            raise CommandLineError(
                "Expecting a STATION, got {}".format(stnName)
            )
        cmdenv.system = station.system.name()
        cmdenv.station = station.dbname
        
        return station.system, station
    
    # Clean up the station name and potentially lift the system
    # name out of it.
    stnName = re.sub(r" +", " ", stnName)
    stnName = re.sub(r"[ /]*/[ /]*", "/", stnName)
    while stnName.startswith('/'):
        stnName = stnName[1:]
    slashPos = stnName.find('/')
    if slashPos > 0:
        sysName, stnName = stnName[:slashPos], stnName[slashPos+1:]
        sysName = sysName.upper()
    else:
        sysName = None
    
    if not stnName:
        raise CommandLineError("Invalid station name: {}".format(
                envStnName
        ))
    
    if not sysName:
        raise CommandLineError("No system name specified")
    
    cmdenv.system, cmdenv.station = sysName, utils.titleFixup(stnName)
    try:
        system = tdb.lookupSystem(sysName)
    except LookupError:
        raise CommandLineError(
                "Unknown SYSTEM name: \"{}\"".format(
                    sysName
        ))
    
    # check the station does not exist
    checkStationDoesNotExist(tdb, cmdenv, system, stnName)
    
    return system, None


def addStation(tdb, cmdenv, system, stationName):
    return tdb.addLocalStation(
            system=system,
            name=stationName,
            lsFromStar=cmdenv.lsFromStar or 0,
            market=cmdenv.market or '?',
            blackMarket=cmdenv.blackMarket or '?',
            shipyard=cmdenv.shipyard or '?',
            outfitting=cmdenv.outfitting or '?',
            rearm=cmdenv.rearm or '?',
            refuel=cmdenv.refuel or '?',
            repair=cmdenv.repair or '?',
            maxPadSize=cmdenv.padSize or '?',
            planetary=cmdenv.planetary or '?',
            commit=True,
    )


def updateStation(tdb, cmdenv, station):
    return tdb.updateLocalStation(
            station=station,
            lsFromStar=cmdenv.lsFromStar,
            market=cmdenv.market,
            blackMarket=cmdenv.blackMarket,
            shipyard=cmdenv.shipyard,
            outfitting=cmdenv.outfitting,
            rearm=cmdenv.rearm,
            refuel=cmdenv.refuel,
            repair=cmdenv.repair,
            maxPadSize=cmdenv.padSize,
            planetary=cmdenv.planetary,
            force=True,
            commit=True,
    )


def removeStation(tdb, cmdenv, station):
    db = tdb.getDB()
    db.execute("""
            DELETE FROM Station WHERE station_id = ?
    """, [station.ID])
    db.commit()
    cmdenv.NOTE("{} (#{}) removed from {} database.",
            station.name(), station.ID, tdb.dbPath)
    cmdenv.stationItemCount = station.itemCount
    return True


def checkResultAndExportStations(tdb, cmdenv, result):
    if not result:
        cmdenv.NOTE("No changes.")
        return None
    if cmdenv.noExport:
        cmdenv.DEBUG0("no-export set, not exporting stations")
        return None
    
    lines, csvPath = csvexport.exportTableToFile(tdb, cmdenv, "Station")
    cmdenv.NOTE("{} updated.", csvPath)
    
    if cmdenv.remove:
        if cmdenv.stationItemCount:
            cmdenv.NOTE("Station had items, regenerating .prices file")
            cache.regeneratePricesFile(tdb, cmdenv)
    
    return None


######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    if cmdenv.lsFromStar and cmdenv.lsFromStar < 0:
        raise CommandLineError("Invalid (negative) --ls option")
    
    system, station = checkSystemAndStation(tdb, cmdenv)
    
    systemName = cmdenv.system
    stationName = cmdenv.station
    
    if cmdenv.add:
        result = addStation(tdb, cmdenv, system, stationName)
        return checkResultAndExportStations(tdb, cmdenv, result)
    elif cmdenv.update:
        result = updateStation(tdb, cmdenv, station)
        return checkResultAndExportStations(tdb, cmdenv, result)
    elif cmdenv.remove:
        result = removeStation(tdb, cmdenv, station)
        return checkResultAndExportStations(tdb, cmdenv, result)
    
    # Otherwise, it's just a query
    results.summary = ResultRow()
    results.summary.system = station.system
    results.summary.station = station
    
    avgSell = results.summary.avgSelling = tdb.getAverageSelling()
    avgBuy = results.summary.avgBuying = tdb.getAverageBuying()
    
    class ItemTrade(object):
        def __init__(self, ID, price, avgAgainst):
            self.ID, self.item = ID, tdb.itemByID[ID]
            self.price = int(price)
            self.avgTrade = avgAgainst.get(ID, 0)
    
    # Look up all selling and buying by the station
    selling, buying = [], []
    cur = tdb.query("""
        SELECT  item_id, demand_price, supply_price
          FROM  StationItem
         WHERE  station_id = ?
                AND (demand_price > 10 or supply_price > 10)
    """, [station.ID])
    for ID, demand_price, supply_price in cur:
        if demand_price > 10 and avgSell.get(ID, 0) > 10:
            buying.append(ItemTrade(ID, demand_price, avgSell))
        if supply_price > 10 and avgBuy.get(ID, 0) > 10:
            selling.append(ItemTrade(ID, supply_price, avgBuy))
    selling.sort(
            key=lambda item: item.price - item.avgTrade,
    )
    results.summary.selling = selling[:5]
    buying.sort(
            key=lambda item: item.avgTrade - item.price,
    )
    results.summary.buying = buying[:5]
    
    return results

def render(results, cmdenv, tdb):
    system, station = results.summary.system, results.summary.station
    
    if cmdenv.detail:
        sysDetail = "(#{} @ {},{},{})".format(
            system.ID, system.posX, system.posY, system.posZ
        )
    else:
        sysDetail = "(#{})".format(system.ID)
    
    print("Station Data:")
    print("System....:", system.name(), sysDetail)
    print("Station...:", station.dbname, "(#{})".format(station.ID))
    
    if cmdenv.detail:
        siblings = ", ".join(
            stn.dbname
            for stn in system.stations
            if stn is not station
        )
        if siblings:
            print("Also Here.:", siblings)
    
    ls = station.distFromStar()
    if cmdenv.detail and ls == '?':
        ls = '0 [unknown]'
    print("Stn/Ls....:", ls)
    
    def _detail(value, source):
        detail = source[value]
        if cmdenv.detail and detail == '?':
            detail += ' [unknown]'
        return detail
    print("Pad Size..:", _detail(station.maxPadSize, TradeDB.padSizes))
    print("Market....:", _detail(station.market, TradeDB.marketStates))
    print("B/Market..:", _detail(station.blackMarket, TradeDB.marketStates))
    print("Shipyard..:", _detail(station.shipyard, TradeDB.marketStates))
    print("Outfitting:", _detail(station.outfitting, TradeDB.marketStates))
    print("Rearm.....:", _detail(station.rearm, TradeDB.marketStates))
    print("Refuel....:", _detail(station.refuel, TradeDB.marketStates))
    print("Repair....:", _detail(station.repair, TradeDB.marketStates))
    print("Planetary.:", _detail(station.planetary, TradeDB.planetStates))
    print("Prices....:", station.itemCount or 'None')
    
    if station.itemCount == 0:
        return
    
    newest, oldest = tdb.query("""
            SELECT JULIANDAY('NOW') - JULIANDAY(MAX(si.modified)),
                   JULIANDAY('NOW') - JULIANDAY(MIN(si.modified))
              FROM StationItem si
             WHERE station_id = ?
    """, [station.ID]).fetchone()
    if newest or oldest:
        # less than a quarter hour difference? ignore?
        if abs(newest - oldest) < (1 / (24 * 4)):
            pricesAge = "{:.2f} days".format(oldest)
        else:
            pricesAge = "{:.2f}-{:.2f} days".format(newest, oldest)
    else:
        pricesAge = "[n/a]"
    
    print("Price Age.:", pricesAge)
    
    def makeBest(rows, explanation, alt, maxLen, starFn):
        if not rows:
            return "[n/a]"
        best = []
        for irow in rows:
            star = '*' if starFn(irow.price, irow.avgTrade) else ''
            best.append([irow, star])
        
        if not cmdenv.detail:
            return ', '.join(irow[0].item.name() + irow[1] for irow in best)
        
        bestText = "("+explanation+")"
        for irow in best:
            bestText += "\n    {:<{len}} @ {:7n}cr (Avg {} {:7n}cr)".format(
                    irow[0].item.name(cmdenv.detail) + irow[1],
                    irow[0].price,
                    alt,
                    irow[0].avgTrade,
                    len=maxLen + 1,
            )
        return bestText


    longestNameLen = max(
        max_len(results.summary.selling, key=lambda row: row.item.name(cmdenv.detail)),
        max_len(results.summary.buying, key=lambda row: row.item.name(cmdenv.detail)),
    )
    print("Best Buy..:", makeBest(
            results.summary.selling, "Buy from this station", "Sell", longestNameLen,
            starFn=lambda price, avgCr: \
                price <= (avgCr * 0.9),
    ))
    print("Best Sale.:", makeBest(
            results.summary.buying, "Sell to this station", "Cost", longestNameLen,
            starFn=lambda price, avgCr: \
                price >= (avgCr * 1.1),
    ))


from __future__ import absolute_import, with_statement, print_function, division, unicode_literals
from .parsing import *
from ..tradedb import TradeDB
from ..tradeexcept import TradeException

import itertools
import math
import sys

######################################################################
# Parser config

name='olddata'
help='Show oldest data in database.'
epilog=None
wantsTradeDB=True
arguments = [
]
switches = [
    ParseArgument('--limit',
            help='Maximum number of results to show',
            default=20,
            type=int,
    ),
    ParseArgument('--near',
            help='Find sellers within jump range of this system.',
            type=str
    ),
    ParseArgument('--ly',
            help='[Requires --near] Systems within this range of --near.',
            default=None,
            dest='maxLyPer',
            metavar='N.NN',
            type=float,
    ),
    ParseArgument('--route',
            help='Sort to shortest path',
            action='store_true',
    ),
    ParseArgument('--min-age',
            help='List data older than this number of days.',
            type=float,
            dest='minAge',
    ),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
    ParseArgument('--ls-max',
        help='Only consider stations upto this many ls from their star.',
        metavar='LS',
        dest='maxLs',
        type=int,
        default=0,
    ),
]

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    from .commandenv import ResultRow
    
    cmdenv = results.cmdenv
    tdb = cmdenv.tdb
    srcSystem = cmdenv.nearSystem
    
    results.summary = ResultRow()
    results.limit = cmdenv.limit
    
    fields = [
        "si.station_id",
        "JULIANDAY('NOW') - JULIANDAY(MAX(si.modified))",
        "stn.ls_from_star",
    ]
    
    joins = []
    wheres = []
    havings = []
    
    if cmdenv.minAge:
        wheres.append(
            "(JULIANDAY('NOW') - JULIANDAY(si.modified) >= {})"
            .format(cmdenv.minAge)
        )
    
    nearSys = cmdenv.nearSystem
    if nearSys:
        maxLy = cmdenv.maxLyPer or tdb.maxSystemLinkLy
        maxLy2 = maxLy ** 2
        fields.append(
                "dist2("
                    "sys.pos_x, sys.pos_y, sys.pos_z,"
                    "{}, {}, {}"
                ") AS d2".format(
                    nearSys.posX,
                    nearSys.posY,
                    nearSys.posZ,
        ))
        joins.append("INNER JOIN System sys USING (system_id)")
        wheres.append("""(
                sys.pos_x BETWEEN {} and {}
                AND sys.pos_y BETWEEN {} and {}
                AND sys.pos_z BETWEEN {} and {}
        )""".format(
                nearSys.posX - maxLy,
                nearSys.posX + maxLy,
                nearSys.posY - maxLy,
                nearSys.posY + maxLy,
                nearSys.posZ - maxLy,
                nearSys.posZ + maxLy,
        ))
        havings.append("d2 <= {}".format(maxLy2))
    else:
        fields.append("0")
    
    fieldStr = ','.join(fields)
    
    if joins:
        joinStr = ' '.join(joins)
    else:
        joinStr = ''
    
    if wheres:
        whereStr = 'WHERE ' + ' AND '.join(wheres)
    else:
        whereStr = ''
    
    if havings:
        haveStr = 'HAVING ' + ' AND '.join(havings)
    else:
        haveStr = ''
    
    stmt = """
            SELECT  {fields}
              FROM  StationItem as si
                    INNER JOIN Station stn USING (station_id)
                    {joins}
             {wheres}
             GROUP  BY 1
             {having}
             ORDER  BY 2 DESC
    """.format(
            fields=fieldStr,
            joins=joinStr,
            wheres=whereStr,
            having=haveStr,
    )
    
    cmdenv.DEBUG1(stmt)
    
    padSize = cmdenv.padSize
    planetary = cmdenv.planetary
    noPlanet = cmdenv.noPlanet
    mls = cmdenv.maxLs
    
    for (stnID, age, ls, dist2) in tdb.query(stmt):
        cmdenv.DEBUG2("{}:{}:{}", stnID, age, ls)
        row = ResultRow()
        row.station = tdb.stationByID[stnID]
        row.age = age
        if ls:
            row.ls = "{:n}".format(ls)
        else:
            row.ls = "?"
        row.dist = dist2 ** 0.5
        if not padSize or row.station.checkPadSize(padSize):
            if not planetary or row.station.checkPlanetary(planetary):
                if not noPlanet or row.station.planetary == 'N':
                    if not mls or row.station.lsFromStar <= mls:
                        results.rows.append(row)
    
    if cmdenv.route and len(results.rows) > 1:
        def walk(startNode, dist):
            rows = results.rows
            startNode = rows[startNode]
            openList = set(rows)
            path = [startNode]
            openList.remove(startNode)
            while len(path) < len(rows):
                lastNode = path[-1]
                distFn = lastNode.station.system.distanceTo
                nearest = min(openList, key=lambda row: distFn(row.station.system))
                openList.remove(nearest)
                path.append(nearest)
                dist += distFn(nearest.station.system)
            return (path, dist)
        if cmdenv.near:
            bestPath = walk(0, results.rows[0].dist)
        else:
            bestPath = (results.rows, float("inf"))
            for i in range(len(results.rows)):
                path = walk(i, 0)
                if path[1] < bestPath[1]:
                    bestPath = path
        results.rows[:] = bestPath[0]
    
    if cmdenv.limit:
        results.rows[:] = results.rows[:cmdenv.limit]
    
    return results

######################################################################
# Transform result set into output

def render(results, cmdenv, tdb):
    from ..formatting import RowFormat, ColumnFormat
    
    if not results or not results.rows:
        raise TradeException("No data found")
    
    # Compare system names so we can tell
    longestNamed = max(results.rows,
                    key=lambda row: len(row.station.name()))
    longestNameLen = len(longestNamed.station.name())
    
    rowFmt = RowFormat().append(
            ColumnFormat("Station", '<', longestNameLen,
                    key=lambda row: row.station.name())
    )
    
    if cmdenv.quiet < 2:
        if cmdenv.nearSystem:
            rowFmt.addColumn('DistLy', '>', 6, '.2f',
                    key=lambda row: row.dist
            )
        
        rowFmt.append(
                ColumnFormat("Age/days", '>', '8', '.2f',
                        key=lambda row: row.age)
        ).append(
                ColumnFormat("StnLs", '>', '10',
                        key=lambda row: row.station.distFromStar())
        ).append(
                ColumnFormat("Pad", '>', '3',
                        key=lambda row: \
                            TradeDB.padSizes[row.station.maxPadSize])
        ).append(
                ColumnFormat("Plt", '>', '3',
                        key=lambda row: \
                            TradeDB.planetStates[row.station.planetary]))
    
    if not cmdenv.quiet:
        heading, underline = rowFmt.heading()
        print(heading, underline, sep='\n')
    
    for row in results.rows:
        print(rowFmt.format(row))